from typing import Any, Dict, Optional

from requests.models import Response


class Error(Exception):
    """
    Base class for all API errors
    """

    def __init__(
        self,
        message: Optional[str] = None,
        response: Optional[Response] = None,
        payload: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.message = message
        self.response = response
        self.payload = payload

    def __str__(self) -> str:
        if self.response:
            if not self.message:
                if self.response.status_code >= 500:
                    self.message = "Service unavailable, please try again later"
                if self.response.status_code == 400:
                    self.message = "Please check your inputs and try again"

            return f"Status code {self.response.status_code}. Error message: {self.message}"

        return f"Error message: {self.message}"


class AuthenticationError(Error):
    """
    Raised when an API request fails due to an authentication error
    """

    pass


class APIError(Error):
    """
    Raised when an API request fails due to an unknown error
    """

    pass



from typing import Literal

WorkCardMovementType = Literal["ARRIVAL", "DEPARTURE"]

LateDeclarationJustificationType = Literal[
    "POWER_OUTAGE", "EMPLOYER_SYSTEMS_UNAVAILABLE", "ERGANI_SYSTEMS_UNAVAILABLE"
]

OvertimeJustificationType = Literal[
    "ACCIDENT_PREVENTION_OR_DAMAGE_RESTORATION",
    "URGENT_SEASONAL_TASKS",
    "EXCEPTIONAL_WORKLOAD",
    "SUPPLEMENTARY_TASKS",
    "LOST_HOURS_SUDDEN_CAUSES",
    "LOST_HOURS_OFFICIAL_HOLIDAYS",
    "LOST_HOURS_WEATHER_CONDITIONS",
    "EMERGENCY_CLOSURE_DAY",
    "NON_WORKDAY_TASKS",
]

ScheduleWorkType = Literal["WORK_FROM_OFFICE", "WORK_FROM_HOME", "REST_DAY", "ABSENT"]



from typing import Optional

import requests
from requests.auth import AuthBase
from requests.models import PreparedRequest

from ergani.exceptions import AuthenticationError
from ergani.utils import extract_error_message


class ErganiAuthentication(AuthBase):
    """
    Authentication handler for the Ergani API
    """

    def __init__(
        self,
        username: str,
        password: str,
        base_url: Optional[str] = "https://trialeservices.yeka.gr/WebServicesAPI/api",
    ) -> None:
        self.username = username
        self.password = password
        self.base_url = base_url
        self.access_token = self._authenticate()

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        request.headers["Authorization"] = f"Bearer {self.access_token}"
        return request

    def _authenticate(self) -> str:
        endpoint = "/Authentication"
        payload = {
            "Username": self.username,
            "Password": self.password,
            "UserType": "01",
        }

        response = requests.post(f"{self.base_url}{endpoint}", json=payload)

        if response.status_code != 200:
            error_message = extract_error_message(response)
            raise AuthenticationError(message=error_message, response=response)

        token = response.json()["accessToken"]
        return token



import json
from datetime import date, datetime, time
from typing import Optional, Union

from requests.models import Response

from ergani.typings import (
    LateDeclarationJustificationType,
    OvertimeJustificationType,
    ScheduleWorkType,
    WorkCardMovementType,
)


def extract_error_message(response: Response) -> str:
    """
    Extracts the error message from a requests Response object

    Args:
        response: A Response object

    Returns:
        The extracted error message
    """
    content_type: str = response.headers.get("Content-Type", "")

    if "application/json" in content_type:
        try:
            response_data = response.json()
            if "message" in response_data:
                return response_data["message"]
            if "msg" in response_data:
                return response_data["msg"]
            if "detail" in response_data:
                return response_data["detail"]
            if not response_data:
                return ""
            return response_data
        except json.JSONDecodeError:
            pass

    if "text/plain" in content_type:
        return response.text.strip()

    return ""


def format_time(t: time) -> str:
    """
    Formats a datetime.time instance to `HH:MM`

    Args:
        t: The time that is going to be formatted

    Returns:
        The formatted time
    """

    if not t:
        return ""

    return t.strftime("%H:%M")


def format_date(d: Optional[date]) -> str:
    """
    Formats a datetime.date instance to `dd/nm/YYYY"`

    Args:
        d: The date that is going to be formatted

    Returns:
        The formatted date
    """

    if not d:
        return ""

    return d.strftime("%d/%m/%Y")


def format_datetime(d: Optional[datetime]) -> str:
    """
    Formats a datetime.datetime instance to an ISO 8601 format

    Args:
        d: The datetime that is going to be formatted

    Returns:
        The formatted datetime
    """

    if not d:
        return ""

    return d.strftime("%Y-%m-%dT%H:%M:%S.%f%z")


def get_day_of_week(d: Optional[date]) -> Union[int, str]:
    """
    Returns the day of the week from a datetime.date instance

    0 - Sunday, 6 - Saturday

    Args:
        d: The date that is going to be evaluated

    Returns:
        The day of the week
    """

    if not d:
        return ""

    day_of_week = d.weekday()
    return (day_of_week + 1) % 7


def get_ergani_workcard_movement_type(movement_type: WorkCardMovementType) -> str:
    """
    Returns the ergani value for workcard_movement_type

    Args:
        workcard_movement_type: The movement type of a work card

    Returns:
        The Ergani value that is going to be used with the Ergani API
    """

    movement_type_mapping = {
        "ARRIVAL": "0",
        "DEPARTURE": "1",
    }

    return movement_type_mapping[movement_type]


def get_ergani_late_declaration_justification(
    justification: LateDeclarationJustificationType,
) -> str:
    """
    Returns the ergani value for late_declaration_justification

    Args:
        late_declaration_justification: The justification for a late work card submission

    Returns:
        The Ergani value that is going to be used with the Ergani API
    """

    justification_mapping = {
        "POWER_OUTAGE": "001",
        "EMPLOYER_SYSTEMS_UNAVAILABLE": "002",
        "ERGANI_SYSTEMS_UNAVAILABLE": "003",
    }

    return justification_mapping.get(justification, justification)


def get_ergani_overtime_cancellation(cancellation: bool) -> str:
    """
    Returns the ergani value for overtime_cancellation

    Args:
        overtime_cancellation: A value representing if an overtime is going to to be cancelled or not

    Returns:
        The Ergani value that is going to be used with the Ergani API
    """

    return "0" if not cancellation else "1"


def get_ergani_overtime_justification(justification: OvertimeJustificationType) -> str:
    """
    Returns the ergani value for overtime_justification

    Args:
        overtime_justification: The justification of an employee's overtime

    Returns:
        The Ergani value that is going to be used with the Ergani API
    """

    justification_mapping = {
        "ACCIDENT_PREVENTION_OR_DAMAGE_RESTORATION": "001",
        "URGENT_SEASONAL_TASKS": "002",
        "EXCEPTIONAL_WORKLOAD": "003",
        "SUPPLEMENTARY_TASKS": "004",
        "LOST_HOURS_SUDDEN_CAUSES": "005",
        "LOST_HOURS_OFFICIAL_HOLIDAYS": "006",
        "LOST_HOURS_WEATHER_CONDITIONS": "007",
        "EMERGENCY_CLOSURE_DAY": "008",
        "NON_WORKDAY_TASKS": "009",
    }

    return justification_mapping[justification]


def get_ergani_work_type(work_type: ScheduleWorkType) -> str:
    """
    Returns the ergani value for work_type

    Args:
        work_type: The work type of an employee's schedule

    Returns:
        The Ergani value that is going to be used with the Ergani API
    """

    work_type_mapping = {
        "WORK_FROM_OFFICE": "ΕΡΓ",
        "WORK_FROM_HOME": "ΤΗΛ",
        "REST_DAY": "ΑΝ",
        "ABSENT": "ΜΕ",
    }

    return work_type_mapping[work_type]



from dataclasses import dataclass, field
from datetime import date, datetime, time
from typing import List, Literal, Optional, TypedDict

from ergani.typings import (
    LateDeclarationJustificationType,
    OvertimeJustificationType,
    ScheduleWorkType,
    WorkCardMovementType,
)
from ergani.utils import (
    format_date,
    format_datetime,
    format_time,
    get_day_of_week,
    get_ergani_late_declaration_justification,
    get_ergani_overtime_cancellation,
    get_ergani_overtime_justification,
    get_ergani_work_type,
    get_ergani_workcard_movement_type,
)


@dataclass
class WorkCard:
    """
    Represents a work card entry for an employee

    Attributes:
        employee_tax_identification_number (str): The employee's tax identification number
        employee_last_name (str): The last name of the employee
        employee_first_name (str): The first name of the employee
        work_card_movement_type (WorkCardMovementType): The type of work card movement
        work_card_submission_date (date): The date the work card was submitted
        work_card_movement_datetime (datetime): The exact date and time of the work card movement
        late_declaration_justification (Optional[LateDeclarationJustificationType]): The justification for the late declaration of the work card movement
    """

    employee_tax_identification_number: str
    employee_last_name: str
    employee_first_name: str
    work_card_movement_type: WorkCardMovementType
    work_card_submission_date: date
    work_card_movement_datetime: datetime
    late_declaration_justification: Optional[LateDeclarationJustificationType] = None

    def serialize(self):
        return {
            "f_afm": self.employee_tax_identification_number,
            "f_eponymo": self.employee_last_name,
            "f_onoma": self.employee_first_name,
            "f_type": get_ergani_workcard_movement_type(self.work_card_movement_type),
            "f_reference_date": self.work_card_submission_date.isoformat(),
            "f_date": format_datetime(self.work_card_movement_datetime),
            "f_aitiologia": get_ergani_late_declaration_justification(
                self.late_declaration_justification
            ),
        }


@dataclass
class CompanyWorkCard:
    """
    Represents work card entries that are issued on a single business branch

    Attributes:
        employer_tax_identification_number (str): The employer's tax identification number
        business_branch_number (int): The number identifying the specific business branch
        comments (Optional[str]): Additional comments related to the work cards
        card_details (List[WorkCard]): A list of `WorkCard` entries for the business branch
    """

    employer_tax_identification_number: str
    business_branch_number: int
    comments: Optional[str] = ""
    card_details: List[WorkCard] = field(default_factory=list)

    def serialize(self):
        return {
            "f_afm_ergodoti": self.employer_tax_identification_number,
            "f_aa": self.business_branch_number,
            "f_comments": self.comments,
            "Details": {
                "CardDetails": [
                    work_card.serialize() for work_card in self.card_details
                ]
            },
        }


@dataclass
class Overtime:
    """
    Represents an overtime entry for an employee

    Attributes:
        employee_tax_identification_number (str): The employee's tax identification number
        employee_social_security_number (str): The employee's social security number
        employee_last_name (str): The last name of the employee
        employee_first_name (str): The first name of the employee
        overtime_date (date): The date of the overtime
        overtime_start_time (time): The start time of the overtime period
        overtime_end_time (time): The end time of the overtime period
        overtime_cancellation (bool): Indicates if the overtime was cancelled or not
        employee_profession_code (str): The profession code of the employee
        overtime_justification (OvertimeJustificationType): The justification for the overtime
        weekly_workdays_number (Literal[5, 6]): The number of the employee's working days in a week
        asee_approval (Optional[str]): The ASEE aproval
    """

    employee_tax_identification_number: str
    employee_social_security_number: str
    employee_last_name: str
    employee_first_name: str
    overtime_date: date
    overtime_start_time: time
    overtime_end_time: time
    overtime_cancellation: bool
    employee_profession_code: str
    overtime_justification: OvertimeJustificationType
    weekly_workdays_number: Literal[5, 6]
    asee_approval: Optional[str] = ""

    def serialize(self):
        return {
            "f_afm": self.employee_tax_identification_number,
            "f_amka": self.employee_social_security_number,
            "f_eponymo": self.employee_last_name,
            "f_onoma": self.employee_first_name,
            "f_date": format_date(self.overtime_date),
            "f_from": format_time(self.overtime_start_time),
            "f_to": format_time(self.overtime_end_time),
            "f_cancellation": get_ergani_overtime_cancellation(
                self.overtime_cancellation
            ),
            "f_step": self.employee_profession_code,
            "f_reason": get_ergani_overtime_justification(self.overtime_justification),
            "f_weekdates": self.weekly_workdays_number,
            "f_asee": self.asee_approval,
        }


@dataclass
class CompanyOvertime:
    """
    Represents overtime entries that are issued on a single business branch

    Attributes:
        business_branch_number (int): The number identifying the specific business branch
        sepe_service_code (str): The SEPE service code
        business_primary_activity_code (str): The primary activity code of the business
        business_branch_activity_code (str): The activity code for the specific branch
        kallikratis_municipal_code (str): The kallikratis municipal code
        legal_representative_tax_identification_number (str): Tax identification number of the legal representative
        employee_overtimes (List[Overtime]): A list of `Overtime` entries for employees
        related_protocol_id (Optional[str]): Related protocol ID
        related_protocol_date (Optional[date]): The date of the related protocol
        employer_organization (Optional[str]): The employer's organization name
        business_secondary_activity_code_1 (Optional[str]): Secondary activity code 1
        business_secondary_activity_code_2 (Optional[str]): Secondary activity code 2
        business_secondary_activity_code_3 (Optional[str]): Secondary activity code 3
        business_secondary_activity_code_4 (Optional[str]): Secondary activity code 4
        comments (Optional[str]): Additional comments related to the overtime entries
    """

    business_branch_number: int
    sepe_service_code: str
    business_primary_activity_code: str
    business_branch_activity_code: str
    kallikratis_municipal_code: str
    legal_representative_tax_identification_number: str
    employee_overtimes: List[Overtime] = field(default_factory=list)
    related_protocol_id: Optional[str] = ""
    related_protocol_date: Optional[date] = None
    employer_organization: Optional[str] = ""
    business_secondary_activity_code_1: Optional[str] = ""
    business_secondary_activity_code_2: Optional[str] = ""
    business_secondary_activity_code_3: Optional[str] = ""
    business_secondary_activity_code_4: Optional[str] = ""
    comments: Optional[str] = ""

    def serialize(self):
        return {
            "f_aa_pararthmatos": self.business_branch_number,
            "f_rel_protocol": self.related_protocol_id,
            "f_rel_date": format_date(self.related_protocol_date),
            "f_ypiresia_sepe": self.sepe_service_code,
            "f_ergodotikh_organwsh": self.employer_organization,
            "f_kad_kyria": self.business_primary_activity_code,
            "f_kad_deyt_1": self.business_secondary_activity_code_1,
            "f_kad_deyt_2": self.business_secondary_activity_code_2,
            "f_kad_deyt_3": self.business_secondary_activity_code_3,
            "f_kad_deyt_4": self.business_secondary_activity_code_4,
            "f_kad_pararthmatos": self.business_branch_activity_code,
            "f_kallikratis_pararthmatos": self.kallikratis_municipal_code,
            "f_comments": self.comments,
            "f_afm_proswpoy": self.legal_representative_tax_identification_number,
            "Ergazomenoi": {
                "OvertimeErgazomenosDate": [
                    overtime.serialize() for overtime in self.employee_overtimes
                ]
            },
        }


@dataclass
class WorkdayDetails:
    """
    Represents details of an employee's workday

    Attributes:
        work_type (ScheduleWorkType): The type of an employee's work schedule
        start_time (time): The start time of the workday
        end_time (time): The end time of the workday
    """

    work_type: ScheduleWorkType
    start_time: time
    end_time: time

    def serialize(self):
        return {
            "f_type": get_ergani_work_type(self.work_type),
            "f_from": format_time(self.start_time),
            "f_to": format_time(self.end_time),
        }


@dataclass
class EmployeeDailySchedule:
    """
    Represents a daily schedule entry for an employee

    Attributes:
        employee_tax_identification_number (str): The employee's tax identification number
        employee_last_name (str): The employee's last name
        employee_first_name (str): The employee's first name
        schedule_date (date): The date of the schedule
        workday_details (List[WorkdayDetails]): A list of workday detail entries for the employee
    """

    employee_tax_identification_number: str
    employee_last_name: str
    employee_first_name: str
    schedule_date: date
    workday_details: List[WorkdayDetails] = field(default_factory=list)

    def serialize(self):
        return {
            "f_afm": self.employee_tax_identification_number,
            "f_eponymo": self.employee_last_name,
            "f_onoma": self.employee_first_name,
            "f_date": format_date(self.schedule_date),
            "ErgazomenosAnalytics": {
                "ErgazomenosWTOAnalytics": [
                    workday_detail.serialize()
                    for workday_detail in self.workday_details
                ]
            },
        }


@dataclass
class CompanyDailySchedule:
    """
    Represents daily schedule entries that are issued on a single business branch

    Attributes:
        business_branch_number (int): The number identifying the business branch
        start_date (Optional[date]): The start date of the schedule
        end_date (Optional[date]): The end date of the schedule period
        employee_schedules (List[EmployeeDailySchedule]): A list of daily schedules for employees
        related_protocol_id (Optional[str]): The ID of the related protocol
        related_protocol_date (Optional[date]): The date of the related protocol
        comments (Optional[str]): Additional comments regarding the daily schedule entries
    """

    business_branch_number: int
    start_date: Optional[date] = None
    end_date: Optional[date] = None
    employee_schedules: List[EmployeeDailySchedule] = field(default_factory=list)
    related_protocol_id: Optional[str] = ""
    related_protocol_date: Optional[date] = None
    comments: Optional[str] = ""

    def serialize(self):
        return {
            "f_aa_pararthmatos": self.business_branch_number,
            "f_rel_protocol": self.related_protocol_id,
            "f_rel_date": format_date(self.related_protocol_date),
            "f_comments": self.comments,
            "f_from_date": format_date(self.start_date),
            "f_to_date": format_date(self.end_date),
            "Ergazomenoi": {
                "ErgazomenoiWTO": [
                    employee_schedule.serialize()
                    for employee_schedule in self.employee_schedules
                ]
            },
        }


@dataclass
class EmployeeWeeklySchedule:
    """
    Represents a weekly schedule entry for an employee

    Attributes:
        employee_tax_identification_number (str): The employee's tax identification number
        employee_last_name (str): The employee's last name
        employee_first_name (str): The employee's first name
        schedule_date (date): The date of the schedule
        workday_details (List[WorkdayDetails]): A list of workday detail entries for the week
    """

    employee_tax_identification_number: str
    employee_last_name: str
    employee_first_name: str
    schedule_date: date
    workday_details: List[WorkdayDetails] = field(default_factory=list)

    def serialize(self):
        return {
            "f_afm": self.employee_tax_identification_number,
            "f_eponymo": self.employee_last_name,
            "f_onoma": self.employee_first_name,
            "f_day": get_day_of_week(self.schedule_date),
            "ErgazomenosAnalytics": {
                "ErgazomenosWTOAnalytics": [
                    workday_detail.serialize()
                    for workday_detail in self.workday_details
                ]
            },
        }


@dataclass
class CompanyWeeklySchedule:
    """
    Represents weekly schedule entries that are issued on a single business branch

    Attributes:
        business_branch_number (int): The number identifying the business branch
        start_date (date): The start date of the weekly schedule
        end_date (date): The end date of the weekly schedule
        employee_schedules (List[EmployeeWeeklySchedule]): A list of weekly schedules for employees
        related_protocol_id (Optional[str]): The ID of the related protocol
        related_protocol_date (Optional[date]): The date of the related protocol
        comments (Optional[str]): Additional comments regarding the weekly schedule entries
    """

    business_branch_number: int
    start_date: date
    end_date: date
    employee_schedules: List[EmployeeWeeklySchedule] = field(default_factory=list)
    related_protocol_id: Optional[str] = ""
    related_protocol_date: Optional[date] = None
    comments: Optional[str] = ""

    def serialize(self):
        return {
            "f_aa_pararthmatos": self.business_branch_number,
            "f_rel_protocol": self.related_protocol_id,
            "f_rel_date": format_date(self.related_protocol_date),
            "f_comments": self.comments,
            "f_from_date": format_date(self.start_date),
            "f_to_date": format_date(self.end_date),
            "Ergazomenoi": {
                "ErgazomenoiWTO": [
                    employee_schedule.serialize()
                    for employee_schedule in self.employee_schedules
                ]
            },
        }


class SubmissionResponse(TypedDict):
    """
    Represents a submission response from the Ergani API

    Attributes:
        submission_id (str): The unique identifier of the submission
        protocol (str): The protocol associated with the submission
        submission_date (datetime): The datetime of the submission
    """

    submission_id: str
    protocol: str
    submission_date: datetime



from datetime import datetime
from typing import Any, Dict, List, Optional

import requests
from requests.models import Response

from ergani.auth import ErganiAuthentication
from ergani.exceptions import APIError, AuthenticationError
from ergani.models import (
    CompanyDailySchedule,
    CompanyOvertime,
    CompanyWeeklySchedule,
    CompanyWorkCard,
    SubmissionResponse,
)
from ergani.utils import extract_error_message


class ErganiClient:
    """
    A client for interacting with the Ergani API

    Args:
        username str: The username for authentication with Ergani
        password str: The password for authentication with Ergani
        base_url str: The base URL of the Ergani API. Defaults to "https://trialeservices.yeka.gr/WebServicesAPI/api".
    """

    def __init__(
        self,
        username: str,
        password: str,
        base_url: Optional[str] = "https://trialeservices.yeka.gr/WebServicesAPI/api",
    ) -> None:
        self.username = username
        self.password = password
        self.base_url = base_url

    def _request(
        self, method: str, endpoint: str, payload: Optional[Dict[str, Any]] = None
    ) -> Optional[Response]:
        """
        Sends a request to the specified endpoint using the given HTTP method and payload

        Args:
            method (str): The HTTP method to use for the request (e.g., 'GET', 'POST')
            endpoint (str): The API endpoint to which the request should be sent to
            payload (Optional[Dict[str, Any]]): The JSON-serializable dictionary to be sent as the request payload

        Returns:
            Optional[Response]: The response object from the requests library. Returns None for 204 No Content responses.

        Raises:
            Requests exceptions may be raised for network-related errors
        """

        url = f"{self.base_url}/{endpoint}"
        auth = ErganiAuthentication(self.username, self.password, self.base_url)

        response = requests.request(
            method,
            url,
            json=payload,
            auth=auth,
        )

        return self._handle_response(response, payload)

    def _handle_response(
        self, response: Response, payload: Optional[Dict[str, Any]] = None
    ) -> Optional[Response]:
        """
        Handles the HTTP response, raising exceptions for error status codes and returning the response for successful ones

        Args:
            response (Response): The response object to handle
            payload (Optional[Dict[str, Any]]): The original request payload for inclusion in exceptions if needed

        Returns:
            Optional[Response]: The original response object for successful requests or None for 204 No Content responses

        Raises:
            APIError: An error occurred while communicating with the Ergani API
            AuthenticationError: Raised if there is an authentication error with the Ergani API
        """

        if response.status_code == 401:
            error_message = extract_error_message(response)
            raise AuthenticationError(message=error_message, response=response)

        if response.status_code == 204:
            return None

        try:
            response.raise_for_status()
            return response
        except:
            error_message = extract_error_message(response)
            raise APIError(message=error_message, response=response, payload=payload)

    def _extract_submission_result(
        self, response: Optional[Response]
    ) -> List[SubmissionResponse]:
        """
        Extracts the submission result from the Ergani API response

        Args:
            response (Response): The response object from the Ergani API

        Returns:
            List[SubmissionResponse]: A list of submission responses parsed from the API response

        Raises:
            ValueError: If the response cannot be parsed into submission responses, indicating an unexpected format
        """

        if not response:
            return []

        data = response.json()
        submissions = []

        for submission in data:
            submission_date_str = submission["submitDate"]
            submission_date = datetime.strptime(submission_date_str, "%d/%m/%Y %H:%M")

            submission_response = SubmissionResponse(
                submission_id=submission["id"],
                protocol=submission["protocol"],
                sumbmission_date=submission_date,
            )
            submissions.append(submission_response)

        return submissions

    def submit_work_card(
        self, company_work_cards: List[CompanyWorkCard]
    ) -> List[SubmissionResponse]:
        """
        Submits work card records (check-in, check-out) for employees to the Ergani API

        Args:
            company_work_cards List[CompanyWorkCard]: A list of CompanyWorkCard instances to be submitted

        Returns:
            List[SubmissionResponse]: A list of SumbmissionResponse that were parsed from the API response

        Raises:
            APIError: An error occurred while communicating with the Ergani API
            AuthenticationError: Raised if there is an authentication error with the Ergani API
        """

        endpoint = "/Documents/WRKCardSE"

        request_payload = {
            "Cards": {
                "Card": [
                    company_card.serialize() for company_card in company_work_cards
                ]
            }
        }

        response = self._request("POST", endpoint, request_payload)

        return self._extract_submission_result(response)

    def submit_overtime(
        self, company_overtimes: List[CompanyOvertime]
    ) -> List[SubmissionResponse]:
        """
        Submits overtime records for employees to the Ergani API

        Args:
            company_overtimes List[CompanyOvertime]: A list of CompanyOvertime instances to be submitted

        Returns:
            List[SubmissionResponse]: A list of SumbmissionResponse that were parsed from the API response

        Raises:
            APIError: An error occurred while communicating with the Ergani API
            AuthenticationError: Raised if there is an authentication error with the Ergani API
        """

        endpoint = "/Documents/OvTime"

        request_payload = {
            "Overtimes": {
                "Overtime": [
                    company_overtime.serialize()
                    for company_overtime in company_overtimes
                ]
            }
        }

        response = self._request("POST", endpoint, request_payload)

        return self._extract_submission_result(response)

    def submit_daily_schedule(
        self, company_daily_schedules: List[CompanyDailySchedule]
    ) -> List[SubmissionResponse]:
        """
        Submits schedule records that are updated on a daily basis for employees to the Ergani API

        Args:
            company_daily_schedules List[CompanyDailySchedule]: A list of CompanyDailySchedule instances to be submitted

        Returns:
            List[SubmissionResponse]: A list of SumbmissionResponse that were parsed from the API response

        Raises:
            APIError: An error occurred while communicating with the Ergani API
            AuthenticationError: Raised if there is an authentication error with the Ergani API
        """

        endpoint = "/Documents/WTODaily"

        request_payload = {
            "WTOS": {
                "WTO": [schedule.serialize() for schedule in company_daily_schedules]
            }
        }

        response = self._request("POST", endpoint, request_payload)

        return self._extract_submission_result(response)

    def submit_weekly_schedule(
        self, company_weekly_schedules: List[CompanyWeeklySchedule]
    ) -> List[SubmissionResponse]:
        """
        Submits weekly schedule records for employees to the Ergani API

        Args:
            company_weekly_schedules List[CompanyWeeklySchedule]: A list of CompanyWeeklySchedule instances to be submitted

        Returns:
            List[SubmissionResponse]: A list of SumbmissionResponse that were parsed from the API response

        Raises:
            APIError: An error occurred while communicating with the Ergani API
            AuthenticationError: Raised if there is an authentication error with the Ergani API
        """

        endpoint = "/Documents/WTOWeek"

        request_payload = {
            "WTOS": {
                "WTO": [schedule.serialize() for schedule in company_weekly_schedules]
            }
        }

        response = self._request("POST", endpoint, request_payload)

        return self._extract_submission_result(response)
